#include "bbe251.h"
#include "fieldelement.h"

inline void transpose2(__m128i *a)
{
  __m128i x = a[0];
  __m128i y = a[64];
  a[0] = _mm_unpacklo_epi64(x,y);
  a[64] = _mm_unpackhi_epi64(x,y);
}

inline void transpose4(__m128i *a)
{
  int i;
  for (i = 0;i < 128;i += 64) {
    __m128i x = a[i];
    __m128i y = a[i + 32];
    a[i] = _mm_unpacklo_epi32(x,y);
    a[i + 32] = _mm_unpackhi_epi32(x,y);
  }
  transpose2(a);
  transpose2(a + 32);
}

inline void transpose8(__m128i *a)
{
  int i;
  for (i = 0;i < 128;i += 32) {
    __m128i x = a[i];
    __m128i y = a[i + 16];
    a[i] = _mm_unpacklo_epi16(x,y);
    a[i + 16] = _mm_unpackhi_epi16(x,y);
  }
  transpose4(a);
  transpose4(a + 16);
}

void transpose16(__m128i *a)
{
  int i;
  for (i = 0;i < 128;i += 16) {
    __m128i x = a[i];
    __m128i y = a[i + 8];
    a[i] = _mm_unpacklo_epi8(x,y);
    a[i + 8] = _mm_unpackhi_epi8(x,y);
  }
  transpose8(a);
  transpose8(a + 8);
}

void transpose32(__m128i *a)
{
  int i;
  const __m128i lo = _mm_set_epi32(0x0f0f0f0f,0x0f0f0f0f,0x0f0f0f0f,0x0f0f0f0f);
  const __m128i hi = _mm_set_epi32(0xf0f0f0f0,0xf0f0f0f0,0xf0f0f0f0,0xf0f0f0f0);
  for (i = 0;i < 128;i += 8) {
    __m128i x = a[i];
    __m128i y = a[i + 4];
    a[i] = _mm_xor_si128(_mm_and_si128(x,lo),_mm_slli_epi32(_mm_and_si128(y,lo),4));
    a[i + 4] = _mm_xor_si128(_mm_and_si128(y,hi),_mm_srli_epi32(_mm_and_si128(x,hi),4));
  }
  transpose16(a);
  transpose16(a + 4);
}

void transpose64(__m128i *a)
{
  int i;
  const __m128i lo = _mm_set_epi32(0x33333333,0x33333333,0x33333333,0x33333333);
  const __m128i hi = _mm_set_epi32(0xcccccccc,0xcccccccc,0xcccccccc,0xcccccccc);
  for (i = 0;i < 128;i += 4) {
    __m128i x = a[i];
    __m128i y = a[i + 2];
    a[i] = _mm_xor_si128(_mm_and_si128(x,lo),_mm_slli_epi32(_mm_and_si128(y,lo),2));
    a[i + 2] = _mm_xor_si128(_mm_and_si128(y,hi),_mm_srli_epi32(_mm_and_si128(x,hi),2));
  }
  transpose32(a);
  transpose32(a + 2);
}

void transpose(vec (&n)[251],const unsigned char nstr[128][32])
{
  int i;
  int itop;
  int j;
  const __m128i lo = _mm_set_epi32(0x55555555,0x55555555,0x55555555,0x55555555);
  const __m128i hi = _mm_set_epi32(0xaaaaaaaa,0xaaaaaaaa,0xaaaaaaaa,0xaaaaaaaa);
  __m128i r[256];
  for (i = 0;i < 128;i += 2) {
    __m128i x = _mm_loadu_si128((__m128i *) &nstr[i][0]);
    __m128i y = _mm_loadu_si128((__m128i *) &nstr[i + 1][0]);
    r[i] = _mm_xor_si128(_mm_and_si128(x,lo),_mm_slli_epi32(_mm_and_si128(y,lo),1));
    r[i + 1] = _mm_xor_si128(_mm_and_si128(y,hi),_mm_srli_epi32(_mm_and_si128(x,hi),1));
    x = _mm_loadu_si128((__m128i *) &nstr[i][16]);
    y = _mm_loadu_si128((__m128i *) &nstr[i + 1][16]);
    r[i + 128] = _mm_xor_si128(_mm_and_si128(x,lo),_mm_slli_epi32(_mm_and_si128(y,lo),1));
    r[i + 129] = _mm_xor_si128(_mm_and_si128(y,hi),_mm_srli_epi32(_mm_and_si128(x,hi),1));
  }
  transpose64(&r[0]);
  transpose64(&r[1]);
  transpose64(&r[128]);
  transpose64(&r[129]);
  for (itop = 0;itop < 120;itop += 8) {
    j =
      ((itop >> 3) & 8) |
      ((itop >> 1) & 16) |
      ((itop << 1) & 32) |
      ((itop << 3) & 64) ;
    for (i = 0;i < 8;++i) {
      n[itop + i] = r[j + i];
      n[itop + i + 128] = r[i + j + 128];
    }
  }
  for (itop = 120;itop < 128;itop += 8) {
    j =
      ((itop >> 3) & 8) |
      ((itop >> 1) & 16) |
      ((itop << 1) & 32) |
      ((itop << 3) & 64) ;
    for (i = 0;i < 8;++i) {
      n[itop + i] = r[j + i];
      if (itop + i + 128 < 251) n[itop + i + 128] = r[i + j + 128];
    }
  }
}

void transpose(unsigned char (&nstr)[128][32],const vec n[251])
{
  int i;
  int itop;
  int j;

  const __m128i lo = _mm_set_epi32(0x55555555,0x55555555,0x55555555,0x55555555);
  const __m128i hi = _mm_set_epi32(0xaaaaaaaa,0xaaaaaaaa,0xaaaaaaaa,0xaaaaaaaa);
  __m128i r[256];
  for (i = 0;i < 250;i += 2) {
    __m128i x = n[i];
    __m128i y = n[i + 1];
    r[i] = _mm_xor_si128(_mm_and_si128(x,lo),_mm_slli_epi32(_mm_and_si128(y,lo),1));
    r[i + 1] = _mm_xor_si128(_mm_and_si128(y,hi),_mm_srli_epi32(_mm_and_si128(x,hi),1));
  }
  __m128i x = n[i];
  r[i] = _mm_and_si128(x,lo);
  r[i + 1] = _mm_srli_epi32(_mm_and_si128(x,hi),1);
  r[i + 2] = _mm_set_epi32(0,0,0,0);
  r[i + 3] = _mm_set_epi32(0,0,0,0);
  r[i + 4] = _mm_set_epi32(0,0,0,0);
  r[i + 5] = _mm_set_epi32(0,0,0,0);
  transpose64(&r[0]);
  transpose64(&r[1]);
  transpose64(&r[128]);
  transpose64(&r[129]);
  for (itop = 0;itop < 128;itop += 8) {
    j =
      ((itop >> 3) & 8) |
      ((itop >> 1) & 16) |
      ((itop << 1) & 32) |
      ((itop << 3) & 64) ;
    for (i = 0;i < 8;++i) {
      _mm_storeu_si128((__m128i *) &nstr[i + j][0],r[itop + i]);
      _mm_storeu_si128((__m128i *) &nstr[i + j][16],r[itop + i + 128]);
    }
  }
}

void transpose(fieldelement &p,const unsigned char pstr[128][32])
{
  transpose(p.x,pstr);
}

void transpose(unsigned char (&pstr)[128][32],const fieldelement p)
{
  transpose(pstr,p.x);
}

void bbe251(unsigned char (&npstr)[128][32],const unsigned char pstr[128][32],const unsigned char nstr[128][32])
{
  vec n[251];
  fieldelement p;
  fieldelement np;
  transpose(n,nstr);
  transpose(p,pstr);
  np = p.bbe251(n);
  transpose(npstr,np);
}
